"""
Vaex is a library for dealing with larger than memory DataFrames (out of core).

The most important class (datastructure) in vaex is the :class:`.DataFrame`. A DataFrame is obtained by either opening
the example dataset:

>>> import vaex
>>> df = vaex.example()

Or using :func:`open` to open a file.

>>> df1 = vaex.open("somedata.hdf5")
>>> df2 = vaex.open("somedata.fits")
>>> df2 = vaex.open("somedata.arrow")
>>> df4 = vaex.open("somedata.csv")

Or connecting to a remove server:

>>> df_remote = vaex.open("http://try.vaex.io/nyc_taxi_2015")


A few strong features of vaex are:

 * Performance: works with huge tabular data, process over a billion (> 10\\ :sup:`9`\\ ) rows/second.
 * Expression system / Virtual columns: compute on the fly, without wasting ram.
 * Memory efficient: no memory copies when doing filtering/selections/subsets.
 * Visualization: directly supported, a one-liner is often enough.
 * User friendly API: you will only need to deal with a DataFrame object, and tab completion + docstring will help you out: `ds.mean<tab>`, feels very similar to Pandas.
 * Very fast statistics on N dimensional grids such as histograms, running mean, heatmaps.


Follow the tutorial at https://docs.vaex.io/en/latest/tutorial.html to learn how to use vaex.

"""  # -*- coding: utf-8 -*-
from __future__ import print_function
import glob
import re
from numpy.lib.function_base import copy
import six

import vaex.dataframe
import vaex.dataset
from vaex.functions import register_function
from . import stat
# import vaex.file
# import vaex.export
from .delayed import delayed
from .groupby import *
from . import agg
import vaex.datasets
# import vaex.plot
# from vaex.dataframe import DataFrame
# del ServerRest, DataFrame

import vaex.settings
import logging
import pkg_resources
import os
from functools import reduce

try:
    from . import version
except:
    import sys
    print("version file not found, please run git/hooks/post-commit or git/hooks/post-checkout and/or install them as hooks (see git/README)", file=sys.stderr)
    raise

__version__ = version.get_versions()


def app(*args, **kwargs):
    """Create a vaex app, the QApplication mainloop must be started.

    In ipython notebook/jupyter do the following:

    >>> import vaex.ui.main # this causes the qt api level to be set properly
    >>> import vaex

    Next cell:

    >>> %gui qt

    Next cell:

    >>> app = vaex.app()

    From now on, you can run the app along with jupyter

    """

    import vaex.ui.main
    return vaex.ui.main.VaexApp()


def open(path, convert=False, shuffle=False, fs_options={}, fs=None, *args, **kwargs):
    """Open a DataFrame from file given by path.

    Example:

    >>> df = vaex.open('sometable.hdf5')
    >>> df = vaex.open('somedata*.csv', convert='bigdata.hdf5')

    :param str or list path: local or absolute path to file, or glob string, or list of paths
    :param convert: Uses `dataframe.export` when convert is a path. If True, ``convert=path+'.hdf5'``
                    The conversion is skipped if the input file or conversion argument did not change.
    :param bool shuffle: shuffle converted DataFrame or not
    :param dict fs_options: Extra arguments passed to an optional file system if needed:
        * Amazon AWS S3
            * `anonymous` - access file without authentication (public files)
            * `access_key` - AWS access key, if not provided will use the standard env vars, or the `~/.aws/credentials` file 
            * `secret_key` - AWS secret key, similar to `access_key`
            * `profile` - If multiple profiles are present in `~/.aws/credentials`, pick this one instead of 'default', see https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-files.html
            * `region` - AWS Region, e.g. 'us-east-1`, will be determined automatically if not provided.
            * `endpoint_override` - URL/ip to connect to, instead of AWS, e.g. 'localhost:9000' for minio
        * Google Cloud Storage
            * :py:class:`gcsfs.core.GCSFileSystem`
        In addition you can pass the boolean "cache" option.
    :param fs: Apache Arrow FileSystem object, or FSSpec FileSystem object, if specified, fs_options should be empty.
    :param args: extra arguments for file readers that need it
    :param kwargs: extra keyword arguments
    :return: return a DataFrame on success, otherwise None
    :rtype: DataFrame

    Cloud storage support:

    Vaex supports streaming of HDF5 files from Amazon AWS S3 and Google Cloud Storage.
    Files are by default cached in $HOME/.vaex/file-cache/(s3|gs) such that successive access
    is as fast as native disk access.

    The following common fs_options are used for S3 access:

     * anon: Use anonymous access or not (false by default). (Allowed values are: true,True,1,false,False,0)
     * cache: Use the disk cache or not, only set to false if the data should be accessed once. (Allowed values are: true,True,1,false,False,0)

    All fs_options can also be encoded in the file path as a query string.

    Examples:

    >>> df = vaex.open('s3://vaex/taxi/yellow_taxi_2015_f32s.hdf5', fs_options={'anonymous': True})
    >>> df = vaex.open('s3://vaex/taxi/yellow_taxi_2015_f32s.hdf5?anon=true')
    >>> df = vaex.open('s3://mybucket/path/to/file.hdf5', fs_options={'access_key': my_key, 'secret_key': my_secret_key})
    >>> df = vaex.open(f's3://mybucket/path/to/file.hdf5?access_key={{my_key}}&secret_key={{my_secret_key}}')
    >>> df = vaex.open('s3://mybucket/path/to/file.hdf5?profile=myproject')

    Google Cloud Storage support:

    The following fs_options are used for GCP access:

     * token: Authentication method for GCP. Use 'anon' for annonymous access. See https://gcsfs.readthedocs.io/en/latest/index.html#credentials for more details.
     * cache: Use the disk cache or not, only set to false if the data should be accessed once. (Allowed values are: true,True,1,false,False,0).
     * project and other arguments are passed to :py:class:`gcsfs.core.GCSFileSystem`

    Examples:

    >>> df = vaex.open('gs://vaex-data/airlines/us_airline_data_1988_2019.hdf5', fs_options={'token': None})
    >>> df = vaex.open('gs://vaex-data/airlines/us_airline_data_1988_2019.hdf5?token=anon')
    >>> df = vaex.open('gs://vaex-data/testing/xys.hdf5?token=anon&cache=False')
    """
    import vaex
    import vaex.convert
    try:
        path = vaex.file.stringyfy(path)
        if path in aliases:
            path = aliases[path]
        path = vaex.file.stringyfy(path)
        if path.startswith("http://") or path.startswith("ws://") or \
           path.startswith("vaex+http://") or path.startswith("vaex+ws://"):  # TODO: think about https and wss
            server, name = path.rsplit("/", 1)
            url = urlparse(path)
            if '?' in name:
                name = name[:name.index('?')]
            extra_args = {key: values[0] for key, values in parse_qs(url.query).items()}
            if 'token' in extra_args:
                kwargs['token'] = extra_args['token']
            if 'token_trusted' in extra_args:
                kwargs['token_trusted'] = extra_args['token_trusted']
            client = vaex.connect(server, **kwargs)
            return client[name]
        if path.startswith("cluster"):
            import vaex.enterprise.distributed
            return vaex.enterprise.distributed.open(path, *args, **kwargs)
        else:
            import vaex.file
            import glob
            if isinstance(path, str):
                paths = [path]
            else:
                paths = path
            filenames = []
            for path in paths:
                naked_path, options = vaex.file.split_options(path)
                if glob.has_magic(naked_path):
                    filenames.extend(list(sorted(vaex.file.glob(path, **kwargs))))
                else:
                    filenames.append(path)
            df = None
            if len(filenames) == 0:
                raise IOError(f'File pattern did not match anything {path}')
            filename_hdf5 = vaex.convert._convert_name(filenames, shuffle=shuffle)
            filename_hdf5_noshuffle = vaex.convert._convert_name(filenames, shuffle=False)
            if len(filenames) == 1:
                path = filenames[0]
                # # naked_path, _ = vaex.file.split_options(path, fs_options)
                _, ext, _ = vaex.file.split_ext(path)
                if ext == '.csv':  # special case for csv
                    return vaex.from_csv(path, fs_options=fs_options, fs=fs, convert=convert, **kwargs)
                if convert:
                    path_output = convert if isinstance(convert, str) else filename_hdf5
                    vaex.convert.convert(
                        path_input=path, fs_options_input=fs_options, fs_input=fs,
                        path_output=path_output, fs_options_output=fs_options, fs_output=fs,
                        *args, **kwargs
                    )
                    ds = vaex.dataset.open(path_output, fs_options=fs_options, fs=fs, **kwargs)
                else:
                    ds = vaex.dataset.open(path, fs_options=fs_options, fs=fs, **kwargs)
                df = vaex.from_dataset(ds)
                if df is None:
                    if os.path.exists(path):
                        raise IOError('Could not open file: {}, did you install vaex-hdf5? Is the format supported?'.format(path))
            elif len(filenames) > 1:
                if convert not in [True, False]:
                    filename_hdf5 = convert
                else:
                    filename_hdf5 = vaex.convert._convert_name(filenames, shuffle=shuffle)
                if os.path.exists(filename_hdf5) and convert:  # also check mtime
                    df = vaex.open(filename_hdf5)
                else:
                    dfs = []
                    for filename in filenames:
                        dfs.append(vaex.open(filename, convert=bool(convert), shuffle=shuffle, **kwargs))
                    df = vaex.concat(dfs)
                    if convert:
                        if shuffle:
                            df = df.shuffle()
                        df.export_hdf5(filename_hdf5)
                        df = vaex.open(filename_hdf5)

        if df is None:
            raise IOError('Unknown error opening: {}'.format(path))
        return df
    except:
        logging.getLogger("vaex").error("error opening %r" % path)
        raise


def open_many(filenames):
    """Open a list of filenames, and return a DataFrame with all DataFrames concatenated.

    :param list[str] filenames: list of filenames/paths
    :rtype: DataFrame
    """
    dfs = []
    for filename in filenames:
        filename = filename.strip()
        if filename and filename[0] != "#":
            dfs.append(open(filename))
    return concat(dfs)


def from_samp(username=None, password=None):
    """Connect to a SAMP Hub and wait for a single table load event, disconnect, download the table and return the DataFrame.

    Useful if you want to send a single table from say TOPCAT to vaex in a python console or notebook.
    """
    print("Waiting for SAMP message...")
    import vaex.samp
    t = vaex.samp.single_table(username=username, password=password)
    return from_astropy_table(t.to_table())


def from_astropy_table(table):
    """Create a vaex DataFrame from an Astropy Table."""
    from vaex.astro.astropy_table import DatasetAstropyTable
    ds = DatasetAstropyTable(table=table)
    return vaex.dataframe.DataFrameLocal(ds)


def from_dict(data):
    """Create an in memory dataset from a dict with column names as keys and list/numpy-arrays as values

    Example

    >>> data = {'A':[1,2,3],'B':['a','b','c']}
    >>> vaex.from_dict(data)
      #    A    B
      0    1   'a'
      1    2   'b'
      2    3   'c'

    :param data: A dict of {column:[value, value,...]}
    :rtype: DataFrame

    """
    return vaex.from_arrays(**data)


def from_items(*items):
    """Create an in memory DataFrame from numpy arrays, in contrast to from_arrays this keeps the order of columns intact (for Python < 3.6).

    Example

    >>> import vaex, numpy as np
    >>> x = np.arange(5)
    >>> y = x ** 2
    >>> vaex.from_items(('x', x), ('y', y))
      #    x    y
      0    0    0
      1    1    1
      2    2    4
      3    3    9
      4    4   16

    :param items: list of [(name, numpy array), ...]
    :rtype: DataFrame

    """
    return from_dict(dict(items))


def from_arrays(**arrays):
    """Create an in memory DataFrame from numpy arrays.

    Example

    >>> import vaex, numpy as np
    >>> x = np.arange(5)
    >>> y = x ** 2
    >>> vaex.from_arrays(x=x, y=y)
      #    x    y
      0    0    0
      1    1    1
      2    2    4
      3    3    9
      4    4   16
    >>> some_dict = {'x': x, 'y': y}
    >>> vaex.from_arrays(**some_dict)  # in case you have your columns in a dict
      #    x    y
      0    0    0
      1    1    1
      2    2    4
      3    3    9
      4    4   16

    :param arrays: keyword arguments with arrays
    :rtype: DataFrame
    """
    import numpy as np
    import six
    dataset = vaex.dataset.DatasetArrays(arrays)
    return vaex.dataframe.DataFrameLocal(dataset)


def from_arrow_table(table) -> vaex.dataframe.DataFrame:
    """Creates a vaex DataFrame from an arrow Table.

    :param as_numpy: Will lazily cast columns to a NumPy ndarray.
    :rtype: DataFrame
    """
    from vaex.arrow.dataset import from_table
    return from_dataset(from_table(table=table))


def from_arrow_dataset(arrow_dataset) -> vaex.dataframe.DataFrame:
    '''Create a DataFrame from an Apache Arrow dataset'''
    import vaex.arrow.dataset
    return from_dataset(vaex.arrow.dataset.DatasetArrow(arrow_dataset))


def from_dataset(dataset: vaex.dataset.Dataset) -> vaex.dataframe.DataFrame:
    '''Create a Vaex DataFrame from a Vaex Dataset'''
    return vaex.dataframe.DataFrameLocal(dataset)


def from_scalars(**kwargs):
    """Similar to from_arrays, but convenient for a DataFrame of length 1.

    Example:

    >>> import vaex
    >>> df = vaex.from_scalars(x=1, y=2)

    :rtype: DataFrame
    """
    import numpy as np
    return from_arrays(**{k: np.array([v]) for k, v in kwargs.items()})


def from_pandas(df, name="pandas", copy_index=False, index_name="index"):
    """Create an in memory DataFrame from a pandas DataFrame.

    :param: pandas.DataFrame df: Pandas DataFrame
    :param: name: unique for the DataFrame

    >>> import vaex, pandas as pd
    >>> df_pandas = pd.from_csv('test.csv')
    >>> df = vaex.from_pandas(df_pandas)

    :rtype: DataFrame
    """
    import six
    import pandas as pd
    import numpy as np
    import pyarrow as pa
    columns = {}

    def add(name, column):
        values = column.values
        # the first test is to support (partially) pandas 0.23
        if hasattr(pd.core.arrays, 'integer') and isinstance(values, pd.core.arrays.integer.IntegerArray):
            values = np.ma.array(values._data, mask=values._mask)
        elif hasattr(pd.core.arrays, 'StringArray') and isinstance(values, pd.core.arrays.StringArray):
            values = pa.array(values)
        try:
            columns[name] = vaex.dataset.to_supported_array(values)
        except Exception as e:
            print("could not convert column %s, error: %r, will try to convert it to string" % (name, e))
            try:
                values = values.astype("S")
                columns[name] = vaex.dataset.to_supported_array(values)
            except Exception as e:
                print("Giving up column %s, error: %r" % (name, e))
    for name in df.columns:
        add(str(name), df[name])
    if copy_index:
        add(index_name, df.index)
    return from_dict(columns)


def from_ascii(path, seperator=None, names=True, skip_lines=0, skip_after=0, **kwargs):
    """
    Create an in memory DataFrame from an ascii file (whitespace seperated by default).

    >>> ds = vx.from_ascii("table.asc")
    >>> ds = vx.from_ascii("table.csv", seperator=",", names=["x", "y", "z"])

    :param path: file path
    :param seperator: value seperator, by default whitespace, use "," for comma seperated values.
    :param names: If True, the first line is used for the column names, otherwise provide a list of strings with names
    :param skip_lines: skip lines at the start of the file
    :param skip_after: skip lines at the end of the file
    :param kwargs:
    :rtype: DataFrame
    """

    import vaex.ext.readcol as rc
    ds = vaex.dataframe.DataFrameLocal()
    if names not in [True, False]:
        namelist = names
        names = False
    else:
        namelist = None
    data = rc.readcol(path, fsep=seperator, asdict=namelist is None, names=names, skipline=skip_lines, skipafter=skip_after, **kwargs)
    if namelist:
        for name, array in zip(namelist, data.T):
            ds.add_column(name, array)
    else:
        for name, array in data.items():
            ds.add_column(name, array)
    return ds


def from_json(path_or_buffer, orient=None, precise_float=False, lines=False, copy_index=False, **kwargs):
    """ A method to read a JSON file using pandas, and convert to a DataFrame directly.

    :param str path_or_buffer: a valid JSON string or file-like, default: None
    The string could be a URL. Valid URL schemes include http, ftp, s3,
    gcs, and file. For file URLs, a host is expected. For instance, a local
    file could be ``file://localhost/path/to/table.json``
    :param str orient: Indication of expected JSON string format. Allowed values are
    ``split``, ``records``, ``index``, ``columns``, and ``values``.
    :param bool precise_float: Set to enable usage of higher precision (strtod) function when
    decoding string to double values. Default (False) is to use fast but less precise builtin functionality
    :param bool lines: Read the file as a json object per line.

    :rtype: DataFrame
    """
    # Check for unsupported kwargs
    if kwargs.get('typ') == 'series':
        raise ValueError('`typ` must be set to `"frame"`.')
    if kwargs.get('numpy') == True:
        raise ValueError('`numpy` must be set to `False`.')
    if kwargs.get('chunksize') is not None:
        raise ValueError('`chunksize` must be `None`.')

    import pandas as pd
    return from_pandas(pd.read_json(path_or_buffer, orient=orient, precise_float=precise_float, lines=lines, **kwargs),
                       copy_index=copy_index)


def from_csv(filename_or_buffer, copy_index=False, chunk_size=None, convert=False, fs_options={}, fs=None, **kwargs):
    """
    Read a CSV file as a DataFrame, and optionally convert to an hdf5 file.

    :param str or file filename_or_buffer: CSV file path or file-like
    :param bool copy_index: copy index when source is read via Pandas
    :param int chunk_size: if the CSV file is too big to fit in the memory this parameter can be used to read
        CSV file in chunks. For example:

        >>> import vaex
        >>> for i, df in enumerate(vaex.from_csv('taxi.csv', chunk_size=100_000)):
        >>>     df = df[df.passenger_count < 6]
        >>>     df.export_hdf5(f'taxi_{i:02}.hdf5')

    :param bool or str convert: convert files to an hdf5 file for optimization, can also be a path. The CSV
        file will be read in chunks: either using the provided chunk_size argument, or a default size. Each chunk will
        be saved as a separate hdf5 file, then all of them will be combined into one hdf5 file. So for a big CSV file
        you will need at least double of extra space on the disk. Default chunk_size for converting is 5 million rows,
        which corresponds to around 1Gb memory on an example of NYC Taxi dataset.
    :param kwargs: extra keyword arguments, currently passed to Pandas read_csv function, but the implementation might
        change in future versions.
    :returns: DataFrame
    """
    if not convert:
        return _from_csv_read(filename_or_buffer=filename_or_buffer, copy_index=copy_index,
                              fs_options=fs_options, fs=fs, chunk_size=chunk_size, **kwargs)
    else:
        if chunk_size is None:
            # make it memory efficient by default
            chunk_size = 5_000_000
        import vaex.convert
        path_output = convert if isinstance(convert, str) else vaex.convert._convert_name(filename_or_buffer)
        vaex.convert.convert_csv(
            path_input=filename_or_buffer, fs_options_input=fs_options, fs_input=fs,
            path_output=path_output, fs_options_output=fs_options, fs_output=fs,
            chunk_size=chunk_size,
            copy_index=copy_index,
            **kwargs
        )
        return open(path_output, fs_options=fs_options, fs=fs)


def _from_csv_read(filename_or_buffer, copy_index, chunk_size, fs_options={}, fs=None, **kwargs):
    import pandas as pd
    if not chunk_size:
        with vaex.file.open(filename_or_buffer, fs_options=fs_options, fs=fs, for_arrow=True) as f:
            full_df = pd.read_csv(f, **kwargs)
            return from_pandas(full_df, copy_index=copy_index)
    else:
        def iterator():
            chunk_iterator = pd.read_csv(filename_or_buffer, chunksize=chunk_size, **kwargs)
            for chunk_df in chunk_iterator:
                yield from_pandas(chunk_df, copy_index=copy_index)
        return iterator()


def read_csv(filepath_or_buffer, **kwargs):
    '''Alias to from_csv.'''
    return from_csv(filepath_or_buffer, **kwargs)

aliases = vaex.settings.main.auto_store_dict("aliases")

# py2/p3 compatibility
try:
    from urllib.parse import urlparse, parse_qs
except ImportError:
    from urlparse import urlparse, parse_qs


def connect(url, **kwargs):
    """Connect to hostname supporting the vaex web api.

    :param str hostname: hostname or ip address of server
    :rtype: vaex.server.client.Client
    """
    # dispatch to vaex.server package
    from vaex.server import connect
    return connect(url, **kwargs)


def example():
    """Returns an example DataFrame which comes with vaex for testing/learning purposes.

    :rtype: DataFrame
    """
    return vaex.datasets.helmi_de_zeeuw_10percent.fetch()


def zeldovich(dim=2, N=256, n=-2.5, t=None, scale=1, seed=None):
    """Creates a zeldovich DataFrame.
    """
    import vaex.file
    return vaex.file.other.Zeldovich(dim=dim, N=N, n=n, t=t, scale=scale)


def set_log_level_debug():
    """set log level to debug"""
    import logging
    logging.getLogger("vaex").setLevel(logging.DEBUG)


def set_log_level_info():
    """set log level to info"""
    import logging
    logging.getLogger("vaex").setLevel(logging.INFO)


def set_log_level_warning():
    """set log level to warning"""
    import logging
    logging.getLogger("vaex").setLevel(logging.WARNING)


def set_log_level_exception():
    """set log level to exception"""
    import logging
    logging.getLogger("vaex").setLevel(logging.FATAL)


def set_log_level_off():
    """Disabled logging"""
    import logging
    logging.disable(logging.CRITICAL)


format = "%(levelname)s:%(threadName)s:%(name)s:%(message)s"
logging.basicConfig(level=logging.INFO, format=format)
DEBUG_MODE = bool(os.environ.get('VAEX_DEBUG', ''))
if DEBUG_MODE:
    logging.basicConfig(level=logging.DEBUG)
    set_log_level_debug()
else:
    # logging.basicConfig(level=logging.DEBUG)
    set_log_level_warning()

import_script = os.path.expanduser("~/.vaex/vaex_import.py")
if os.path.exists(import_script):
    try:
        with open(import_script) as f:
            code = compile(f.read(), import_script, 'exec')
            exec(code)
    except:
        import traceback
        traceback.print_stack()


logger = logging.getLogger('vaex')


def register_dataframe_accessor(name, cls=None, override=False):
    """Registers a new accessor for a dataframe

    See vaex.geo for an example.
    """
    def wrapper(cls):
        old_value = getattr(vaex.dataframe.DataFrame, name, None)
        if old_value is not None and override is False:
            raise ValueError("DataFrame already has a property/accessor named %r (%r)" % (name, old_value) )

        def get_accessor(self):
            if name in self.__dict__:
                return self.__dict__[name]
            else:
                self.__dict__[name] = cls(self)
            return self.__dict__[name]
        setattr(vaex.dataframe.DataFrame, name, property(get_accessor))
        return cls
    if cls is None:
        return wrapper
    else:
        return wrapper(cls)


for entry in pkg_resources.iter_entry_points(group='vaex.namespace'):
    logger.warning('(DEPRECATED, use vaex.dataframe.accessor) adding vaex namespace: ' + entry.name)
    try:
        add_namespace = entry.load()
        add_namespace()
    except Exception:
        logger.exception('issue loading ' + entry.name)

_lazy_accessors_map = {}


class _lazy_accessor(object):
    def __init__(self, name, scope, loader, lazy_accessors):
        """When adding an accessor geo.cone, scope=='geo', name='cone', scope may be falsy"""
        self.loader = loader
        self.name = name
        self.scope = scope
        self.lazy_accessors = lazy_accessors

    def __call__(self, obj):
        if self.name in obj.__dict__:
            return obj.__dict__[self.name]
        else:
            cls = self.loader()
            accessor = cls(obj)
            obj.__dict__[self.name] = accessor
            fullname = self.name
            if self.scope:
                fullname = self.scope + '.' + self.name
            if fullname in self.lazy_accessors:
                for name, scope, loader, lazy_accessors in self.lazy_accessors[fullname]:
                    assert fullname == scope
                    setattr(cls, name, property(_lazy_accessor(name, scope, loader, lazy_accessors)))
        return obj.__dict__[self.name]


def _add_lazy_accessor(name, loader, target_class=vaex.dataframe.DataFrame):
    """Internal use see tests/internal/accessor_test.py for usage

    This enables us to have df.foo.bar accessors that lazily loads the modules.
    """
    parts = name.split('.')
    if target_class not in _lazy_accessors_map:
        _lazy_accessors_map[target_class] = {}
    lazy_accessors = _lazy_accessors_map[target_class]
    if len(parts) == 1:
        setattr(target_class, parts[0], property(_lazy_accessor(name, None, loader, lazy_accessors)))
    else:
        scope = ".".join(parts[:-1])
        if scope not in lazy_accessors:
            lazy_accessors[scope] = []
        lazy_accessors[scope].append((parts[-1], scope, loader, lazy_accessors))


for entry in pkg_resources.iter_entry_points(group='vaex.dataframe.accessor'):
    logger.debug('adding vaex accessor: ' + entry.name)
    def loader(entry=entry):
        return entry.load()
    _add_lazy_accessor(entry.name, loader)


for entry in pkg_resources.iter_entry_points(group='vaex.expression.accessor'):
    logger.debug('adding vaex expression accessor: ' + entry.name)
    def loader(entry=entry):
        return entry.load()
    _add_lazy_accessor(entry.name, loader, vaex.expression.Expression)


for entry in pkg_resources.iter_entry_points(group='vaex.plugin'):
    if entry.module_name == 'vaex_arrow.opener':
        # if vaex_arrow package is installed, we ignore it
        continue
    logger.debug('adding vaex plugin: ' + entry.name)
    try:
        add_namespace = entry.load()
        add_namespace()
    except Exception:
        logger.exception('issue loading ' + entry.name)


def concat(dfs, resolver='flexible') -> vaex.dataframe.DataFrame:
    '''Concatenate a list of DataFrames.

    :param resolver: How to resolve schema conflicts, see :meth:`DataFrame.concat`.
    '''
    df, *tail = dfs
    return df.concat(*tail, resolver=resolver)

def vrange(start, stop, step=1, dtype='f8'):
    """Creates a virtual column which is the equivalent of numpy.arange, but uses 0 memory"""
    from .column import ColumnVirtualRange
    return ColumnVirtualRange(start, stop, step, dtype)

def string_column(strings):
    import pyarrow as pa
    return pa.array(strings)


def dtype(type):
    '''Creates a Vaex DataType based on a NumPy or Arrow type'''
    return vaex.datatype.DataType(type)

def dtype_of(ar):
    '''Creates a Vaex DataType from a NumPy or Arrow array'''
    if vaex.array_types.is_arrow_array(ar):
        return dtype(ar.type)
    elif vaex.array_types.is_numpy_array(ar) or isinstance(ar, vaex.column.supported_column_types):
        return dtype(ar.dtype)
    else:
        raise TypeError(f'{ar} is not a an Arrow or NumPy array')



import pytest
import pyarrow as pa
import numpy as np
from vaex.datatype import DataType


def test_primitive():
    t1 = DataType(np.dtype('f8'))
    t2 = DataType(np.dtype('f8'))
    assert t1 == t2
    assert t1 == np.dtype('f8')
    # assert np.dtype('f8') == t2  data type not understood


def test_timedelta64():
    t1 = DataType(np.dtype('timedelta64'))
    assert t1 == 'timedelta'
    assert t1 == 'timedelta64'
    assert t1.is_timedelta


@pytest.mark.parametrize('type', [pa.string(), pa.large_string()])
def test_string(type):
    t1 = DataType(type)
    assert t1 == 'string'
    assert t1 == str
    assert not t1.is_float



from __future__ import division, print_function
import logging
import numpy as np
import pyarrow as pa
import vaex.array_types
import vaex.arrow.numpy_dispatch


from .utils import (_ensure_strings_from_expressions,
    _ensure_string_from_expression,
    _ensure_list,
    _is_limit,
    _isnumber,
    _issequence,
    _is_string,
    _parse_reduction,
    _parse_n,
    _normalize_selection_name,
    _normalize,
    _parse_f,
    _expand,
    _expand_shape,
    _expand_limits,
    as_flat_float,
    as_flat_array,
    _split_and_combine_mask)
from .expression import expression_namespace
from vaex.arrow.numpy_dispatch import wrap, unwrap
import vaex.expression

logger = logging.getLogger('vaex.scopes')


class ScopeBase(object):
    def get(self, attr, default=None):  # otherwise pdb crashes during pytest
        if attr == "__tracebackhide__":
            return False
        return default


class UnitScope(ScopeBase):
    def __init__(self, df, value=None):
        self.df = df
        self.value = value

    def __getitem__(self, variable):
        import astropy.units
        if variable in self.df.units:
            unit = self.df.units[variable]
            return (self.value * unit) if self.value is not None else unit
        elif variable in self.df.virtual_columns:
            return eval(self.df.virtual_columns[variable], expression_namespace, self)
        elif variable in self.df.variables:
            return astropy.units.dimensionless_unscaled  # TODO units for variables?
        else:
            raise KeyError("unkown variable %s" % variable)


class _BlockScope(ScopeBase):
    def __init__(self, df, i1, i2, mask=None, **variables):
        """

        :param DataFrameLocal DataFrame: the *local*  DataFrame
        :param i1: start index
        :param i2: end index
        :param values:
        :return:
        """
        self.df = df
        self.i1 = int(i1)
        self.i2 = int(i2)
        self.variables = variables
        self.values = dict(self.variables)
        self.buffers = {}
        self.mask = mask if mask is not None else None

    def move(self, i1, i2):
        length_new = i2 - i1
        length_old = self.i2 - self.i1
        if length_new > length_old:  # old buffers are too small, discard them
            self.buffers = {}
        else:
            for name in list(self.buffers.keys()):
                self.buffers[name] = self.buffers[name][:length_new]
        self.i1 = int(i1)
        self.i2 = int(i2)
        self.values = dict(self.variables)

    def __contains__(self, name):  # otherwise pdb crashes during pytest
        return name in self.buffers  # not sure this should also include varibles, columns and virtual columns

    def _ensure_buffer(self, column):
        if column not in self.buffers:
            logger.debug("creating column for: %s", column)
            self.buffers[column] = np.zeros(self.i2 - self.i1)

    def evaluate(self, expression, out=None):
        if isinstance(expression, vaex.expression.Expression):
            expression = expression.expression
        try:
            # logger.debug("try avoid evaluating: %s", expression)
            result = self[expression]
        except KeyError:
            # logger.debug("no luck, eval: %s", expression)
            # result = ne.evaluate(expression, local_dict=self, out=out)
            # logger.debug("in eval")
            # eval("def f(")
            result = eval(expression, expression_namespace, self)
            self.values[expression] = wrap(result)
            # if out is not None:
            #   out[:] = result
            #   result = out
            # logger.debug("out eval")
        # logger.debug("done with eval of %s", expression)
        result = unwrap(result)
        return result

    def __getitem__(self, variable):
        # logger.debug("get " + variable)
        # return self.df.columns[variable][self.i1:self.i2]
        if variable == 'df':
            return self  # to support df['no!identifier']
        if variable in expression_namespace:
            return expression_namespace[variable]
        try:
            if variable in self.values:
                return self.values[variable]
            elif variable in self.df.columns:
                offset = self.df._index_start
                # if self.df._needs_copy(variable):
                    # self._ensure_buffer(variable)
                    # self.values[variable] = self.buffers[variable] = self.df.columns[variable][self.i1:self.i2].astype(np.float64)
                    # Previously we casted anything to .astype(np.float64), this led to rounding off of int64, when exporting
                    # self.values[variable] = self.df.columns[variable][offset+self.i1:offset+self.i2][:]
                # else:
                values = self.df.columns[variable][offset+self.i1:offset+self.i2]
                if self.mask is not None:
                    # TODO: we may want to put this in array_types
                    if isinstance(values, (pa.Array, pa.ChunkedArray)):
                        values = values.filter(vaex.array_types.to_arrow(self.mask))
                    else:
                        values = values[self.mask]
                self.values[variable] = wrap(values)
            elif variable in list(self.df.virtual_columns.keys()):
                expression = self.df.virtual_columns[variable]
                if isinstance(expression, dict):
                    function = expression['function']
                    arguments = [self.evaluate(k) for k in expression['arguments']]
                    self.values[variable] = wrap(function(*arguments))
                else:
                    # self._ensure_buffer(variable)
                    values = self.evaluate(expression)
                    self.values[variable] = wrap(values)
                    # self.values[variable] = self.buffers[variable]
            elif variable in self.df.functions:
                f = self.df.functions[variable].f
                return vaex.arrow.numpy_dispatch.autowrapper(f)
            if variable not in self.values:
                raise KeyError("Unknown variables or column: %r" % (variable,))

            return self.values[variable]
        except:
            # logger.exception("error in evaluating: %r" % variable)
            raise


class _BlockScopeSelection(ScopeBase):
    def __init__(self, df, i1, i2, selection=None, cache=False, filter_mask=None):
        self.df = df
        self.i1 = i1
        self.i2 = i2
        self.selection = selection
        self.store_in_cache = cache
        self.filter_mask = filter_mask

    def evaluate(self, expression):
        if expression is True:
            expression = "default"
        try:
            expression = _ensure_string_from_expression(expression)
            result = eval(expression, expression_namespace, self)
        except:
            import traceback as tb
            tb.print_stack()
            raise
        result = unwrap(result)
        return result

    def __contains__(self, name):  # otherwise pdb crashes during pytest
        return False

    def __getitem__(self, variable):
        if variable == "__tracebackhide__":  # required for tracebacks
            return False
        # logger.debug("getitem for selection: %s", variable)
        try:
            selection = self.selection
            if selection is None and self.df.has_selection(variable):
                selection = self.df.get_selection(variable)
            # logger.debug("selection for %r: %s %r", variable, selection, self.df.selection_histories)
            key = (self.i1, self.i2)
            if selection:
                assert variable in self.df._selection_masks, "%s mask not found" % (variable, )
                cache = self.df._selection_mask_caches[variable]
                # logger.debug("selection cache: %r" % cache)
                full_mask = self.df._selection_masks[variable]
                selection_in_cache, mask = cache.get(key, (None, None))

                # logger.debug("mask for %r is %r", variable, mask)
                if selection_in_cache == selection:
                    if self.filter_mask is not None:
                        return wrap(mask[self.filter_mask])
                    return wrap(mask)
                # logger.debug("was not cached")
                if variable in self.df.variables:
                    return wrap(self.df.variables[variable])
                mask_values = selection.evaluate(self.df, variable, self.i1, self.i2, self.filter_mask)
                    
                # get a view on a subset of the mask
                sub_mask = full_mask.view(self.i1, self.i2)
                sub_mask_array = np.asarray(sub_mask)
                # and update it
                if self.filter_mask is not None:  # if we have a mask, the selection we evaluated is also filtered
                    sub_mask_array[:] = 0
                    sub_mask_array[:][self.filter_mask] = mask_values
                else:
                    sub_mask_array[:] = mask_values
                # logger.debug("put selection in mask with key %r" % (key,))
                if self.store_in_cache:
                    cache[key] = selection, sub_mask_array
                    # cache[key] = selection, mask_values
                if self.filter_mask is not None:
                    return wrap(sub_mask_array[self.filter_mask])
                else:
                    return wrap(sub_mask_array)
                # return mask_values
            else:
                offset = self.df._index_start
                if variable in expression_namespace:
                    return wrap(expression_namespace[variable])
                elif variable in self.df.columns:
                    values = self.df.columns[variable][offset+self.i1:offset+self.i2]
                    # TODO: we may want to put this in array_types
                    if self.filter_mask is not None:
                        if isinstance(values, (pa.Array, pa.ChunkedArray)):
                            values = values.filter(vaex.array_types.to_arrow(self.filter_mask))
                        else:
                            values = values[self.filter_mask]
                    return wrap(values)
                elif variable in self.df.variables:
                    return self.df.variables[variable]
                elif variable in self.df.virtual_columns:
                    expression = self.df.virtual_columns[variable]
                    # self._ensure_buffer(variable)
                    if expression == variable:
                        raise ValueError(f'Recursion protection: virtual column {variable} refers to itself')
                    values = self.evaluate(expression)  # , out=self.buffers[variable])
                    return wrap(values)
                elif variable in self.df.functions:
                    f = self.df.functions[variable].f
                    return vaex.arrow.numpy_dispatch.autowrapper(f)
                raise KeyError("Unknown variables or column: %r" % (variable,))
        except:
            import traceback as tb
            tb.print_exc()
            logger.exception("error in evaluating: %r" % variable)
            raise




import os
import copyreg


on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    from .superutils import *
    from . import superutils

    ordered_set = tuple([cls for name, cls in vars(superutils).items() if name.startswith('ordered_set')])

    def pickle(x):
        return type(x), (x.extract(), x.count, x.nan_count, x.null_count)
    for cls in ordered_set:
        copyreg.pickle(cls, pickle)


def counter_type_from_dtype(dtype, transient=True):
    from .array_types import is_string_type
    if is_string_type(dtype):
        if transient:
            postfix = 'string'
        else:
            postfix = 'string' # view not support atm
    else:
        postfix = str(dtype)
        if postfix == '>f8':
            postfix = 'float64'
        if postfix == 'double':  # arrow
            postfix = 'float64'
    name = 'counter_' + postfix
    return globals()[name]


def ordered_set_type_from_dtype(dtype, transient=True):
    from .array_types import is_string_type
    if is_string_type(dtype):
        if transient:
            postfix = 'string'
        else:
            postfix = 'string' #  not support atm
    else:
        postfix = str(dtype)
        if postfix == '>f8':
            postfix = 'float64'
    name = 'ordered_set_' + postfix
    return globals()[name]


def index_type_from_dtype(dtype, transient=True, prime_growth=False):
    from .array_types import is_string_type
    if is_string_type(dtype):
        if transient:
            postfix = 'string'
        else:
            postfix = 'string' #  not support atm
    else:
        postfix = str(dtype)
        if postfix == '>f8':
            postfix = 'float64'
    name = 'index_hash_' + postfix
    if prime_growth:
        name += "_prime_growth"
    return globals()[name]



from __future__ import print_function
# -*- coding: utf-8 -*-
import threading
import queue
import sys
import os
import pickle
import traceback
import time
import signal
import multiprocessing
import vaex.misc.progressbar
try:  # py2/3 compatibility
    from io import StringIO
except ImportError:
    from cStringIO import StringIO


pickle_encoding = "hex"
pickle_protocol = 2


def log(*args):
    # print "LOG: " + " ".join([str(k) for k in args])
    pass


def serialize(obj):
    data = pickle.dumps(obj, pickle_protocol)
    rawdata = data.encode(pickle_encoding)
    return rawdata


def deserialize(rawdata):
    data = rawdata.decode(pickle_encoding)
    obj = pickle.loads(data)
    return obj


class InfoThread(threading.Thread):
    def __init__(self, fullsize, executions):
        threading.Thread.__init__(self)
        self.fullsize = fullsize
        self.executions = executions
        self.setDaemon(True)
        # self.

    def n_done(self):
        n = 0
        for execution in self.executions:
            n += len(execution.results)
        return n

    def run(self):
        while 1:
            count = self.n_done()  # self.fullsize - self.taskQueue.qsize()
            print("%d out of %d tasks completed (%5.2f%%)" % (count, self.fullsize, float(count) / self.fullsize * 100))
            time.sleep(0.1)


class Watchdog(threading.Thread):
    def __init__(self, executions):
        threading.Thread.__init__(self)
        # self.fullsize = fullsize
        self.executions = executions
        # self.setDaemon(True):
        self.setDaemon(True)

    def run(self):
        while 1:
            time.sleep(0.5)
            for execution in self.executions:
                print(execution.isAlive(), execution.error)


class InfoThreadProgressBar(InfoThread):
    def __init__(self, fullsize, executions):
        InfoThread.__init__(self, fullsize, executions)
        self.bar = vaex.misc.progressbar.ProgressBar(0, fullsize)

    def run(self):
        done = False
        error = False
        while (not done) and (not error):
            count = self.n_done()  # self.fullsize - self.taskQueue.qsize()
            self.bar.update(count)
            time.sleep(0.1)
            done = count == self.fullsize
            for execution in self.executions:
                if execution.error:
                    error = True


class LocalExecutor(object):
    def __init__(self, thread, function):
        self.thread = thread
        self.function = function

    def stop(self):
        pass

    def init(self):
        pass

    def __call__(self, *args, **kwargs):
        info = "ok"
        exc_info = None
        result = None
        try:
            result = self.function(*args, **kwargs)
        except:
            info = "exception"
            exc_info = traceback.format_exc()
        return info, exc_info, result


class IOExecutor(object):
    def __init__(self, thread, function, init, id_number):
        self.thread = thread
        self.function = function
        self.initf = init
        self.id_number = id_number
        self.createIO()
        self.error = False

    def __call__(self, *args, **kwargs):
        self.log("p: sending execute command")
        # os.close(w) # we are not gonna write
        self.output.write("execute\n")
        self.output.write(serialize((args, kwargs)))
        self.output.write("\n")
        self.output.flush()
        self.log("p: reading response")
        response = self.input.readline().strip()
        # self.log("p: response:", `response`)
        error, exc_info, result = deserialize(response)
        return error, exc_info, result

    def init(self):
        return True

    def stop(self, error=False):
        self.error = error
        self.log("p: sending stop command")
        self.output.write("stop\n")
        self.output.flush()
        self.log("waitpid")
        self.join()
        self.log("waitpid done")
        self.input.close()
        self.output.close()

    def join(self):
        os.waitpid(self.pid, 0)

    def log(self, *args):
        log(self.thread.getName() + ":", " ".join([str(k) for k in args]))


class ForkExecutor(IOExecutor):
    # def __init__(self, thread, function):
    # IOExecutor.__init__(self, thread, function)

    def createIO(self):
        self.log("forking")
        # self.r, self.w = os.pipe()
        # self.output = os.fdopen(self.w, 'w')
        # self.input = os.fdopen(self.r, 'r')
        self.parent_input_p, self.child_output_p = os.pipe()
        self.child_input_p, self.parent_output_p = os.pipe()
        # self.r_c, self.w_c = os.pipe()
        # self.output_p = os.fdopen(self.w_p, 'w')
        # self.input_p = os.fdopen(self.r_p, 'r')
        # self.output_c = os.fdopen(self.w_c, 'w')
        # self.input_c = os.fdopen(self.r_c, 'r')
        self.pid = os.fork()
        if not self.pid:  # we are the child
            # if self.id_number != 0:
            # signal.signal(signal.SIGINT, signal.SIG_IGN)
            self.child_output = os.fdopen(self.child_output_p, 'w')
            self.child_input = os.fdopen(self.child_input_p, 'r')
            self.output = self.child_output
            self.input = self.child_input
            os.close(self.parent_output_p)
            os.close(self.parent_input_p)
            try:
                self.childPart()
            except BaseException:
                print("oops")
            self.input.close()
            self.output.close()
            os._exit(0)
        else:
            self.parent_output = os.fdopen(self.parent_output_p, 'w')
            self.parent_input = os.fdopen(self.parent_input_p, 'r')
            self.output = self.parent_output
            self.input = self.parent_input
            os.close(self.child_output_p)
            os.close(self.child_input_p)
        # else just continue

    def childPart(self):
        done = False
        self.log("c: child")
        if self.initf is not None:
            self.initf(self.id_number)

        while not done:
            self.log("c: waiting for command...")
            command = self.input.readline().strip()
            self.log("command:", command)
            if command == "execute":
                # execute...
                response = self.input.readline().strip()
                self.log("c: args:", repr(response))
                args, kwargs = deserialize(response)
                info = "ok"
                exc_info = None
                result = None
                try:
                    # self.log("c: calling")
                    result = self.function(*args, **kwargs)
                    # self.log("c: result" +result)
                except BaseException:
                    info = "exception"
                    exc_info = traceback.format_exc()
                    # raise "bla"
                    # done = True
                except KeyboardInterrupt:
                    info = "exception"
                    exc_info = traceback.format_exc()
                    # raise "bla"
                    # done = True
                # encode result
                self.log("c: pickling")
                self.output.write(serialize((info, exc_info, result)))
                self.output.write("\n")
                self.output.flush()
                self.log("c: closing")
            elif command == "stop":
                self.log("c: stopping...")
                done = True
                self.log("c: closed, exiting")
            else:
                done = True
                self.log("c: unknown command", repr(command))
                # and exit
                # os._exit(0) # os._exit is used for children

    def join(self):
        os.waitpid(self.pid, 0)


class Execution(threading.Thread):
    def __init__(self, taskQueue, fork, function, init, id_number, args, kwargs):
        self.taskQueue = taskQueue
        self.fork = fork
        self.function = function
        self.init = init
        self.id_number = id_number
        self.args = args
        self.kwargs = kwargs
        self.results = []
        self.done = False
        self.error = False
        threading.Thread.__init__(self)
        if self.fork:
            self.executor = ForkExecutor(self, self.function, self.init, self.id_number)
        else:
            self.executor = LocalExecutor(self, self.function)

    def log(self, *args):
        log(self.getName() + ":", " ".join([str(k) for k in args]))

    def run(self):
        task = None
        if self.executor.init():
            task = None
            self.log("starting")
            try:
                task = self.taskQueue.get(False)
            except queue.Empty:
                self.log("empty at first try")
            while task is not None and self.error is False:
                tasknr, args = task
                args = list(args)
                common_args = list(self.args)
                args.extend(common_args)
                kwargs = dict(self.kwargs)

                info, exc_info, result = self.executor(*args, **kwargs)
                if info == "exception":
                    print(exc_info)
                    self.error = True
                self.log("r: got result")
                self.results.append((tasknr, result))
                try:
                    task = self.taskQueue.get(False)
                except queue.Empty:
                    self.log("empty queue")
                    task = None
            # if not self.error:
            self.executor.stop(self.error)
            self.log("done")
        else:
            self.log("failure starting slave")
            self.error = True
        self.done = True


def countcores():
    stdin, stdout = os.popen2("cat /proc/cpuinfo | grep processor")
    lines = stdout.readlines()
    return len(lines)


def timed(f):
    def execute(*args, **kwargs):
        t0 = time.time()
        utime0, stime0, child_utime0, child_stime0, walltime0 = os.times()
        result = f(*args, **kwargs)
        dt = time.time() - t0
        utime, stime, child_utime, child_stime, walltime = os.times()
        # name = f.func_name
        print()
        print("user time:            % 9.3f sec." % (utime - utime0))
        print("system time:          % 9.3f sec." % (stime - stime0))
        print("user time(children):  % 9.3f sec." % (child_utime - child_utime0))
        print("system time(children):% 9.3f sec." % (child_stime - child_stime0))
        print()
        dt_total = child_utime - child_utime0 + child_stime - child_stime0 + utime - utime0 + stime - stime0
        print("total cpu time:       % 9.3f sec. (time it would take on a single cpu/core)" % (dt_total))
        print("elapsed time:         % 9.3f sec. (normal wallclock time it took)" % (walltime - walltime0))
        dt = walltime - walltime0
        if dt == 0:
            eff = 0.
        else:
            eff = dt_total / (dt)
        print("efficiency factor     % 9.3f      (ratio of the two above ~= # cores)" % eff)
        return result
    return execute


def parallelize(cores=None, fork=True, flatten=False, info=False, infoclass=InfoThreadProgressBar, init=None, *args, **kwargs):
    """Function decorater that executes the function in parallel

    Usage::

            @parallelize(cores=10, info=True)
            def f(x):
                    return x**2

            x = numpy.arange(0, 100, 0.1)
            y = f(x) # this gets executed parallel

    :param cores: number of cpus/cores to use (if None, it counts the cores using /proc/cpuinfo)
    :param fork: fork a process (should always be true since of the GIT, but can be used with c modules that release the GIT)
    :param flatten: if False and each return value is a list, final result will be a list of lists, if True, all lists are combined to one big list
    :param info: show progress bar (see infoclass)
    :param infoclass: class to instantiate that shows the progress (default shows progressbar)
    :param init: function to be called in each forked process before executing, can be used to set the seed, takes a integer as parameter (number that identifies the process)
    :param args: extra arguments passed to function
    :param kwargs: extra keyword arguments passed to function

    Example::

            @parallelize(cores=10, info=True, n=2)
            def f(x, n):
                    return x**n

            x = numpy.arange(0, 100, 0.1)
            y = f(x) # this gets executed parallel



    """
    if cores is None:
        cores = multiprocessing.cpu_count()

    def wrapper(f):
        def execute(*multiargs):
            results = []
            len(list(zip(*multiargs)))
            N = len(multiargs[0])
            if info:
                print("running %i jobs on %i cores" % (N, cores))
            taskQueue = queue.Queue(len(multiargs[0]))
            # for timenr in range(times):
            # taskQueue.put(timenr)
            for tasknr, _args in enumerate(zip(*multiargs)):
                taskQueue.put((tasknr, list(_args)))
            # for timenr in range(times):
            # result = f(*args, **kwargs)
            # results.append(result)
            executions = [Execution(taskQueue, fork, f, init, corenr, args, kwargs) for corenr in range(cores)]
            if info:
                infoobj = infoclass(len(multiargs[0]), executions)
                infoobj.start()
            for i, execution in enumerate(executions):
                execution.setName("T-%d" % i)
                execution.start()
            # if 1:
            # watchdog = Watchdog(executions)
            # watchdog.start()
            error = False
            for execution in executions:
                log("joining:", execution.getName())
                try:
                    execution.join()
                except BaseException:
                    error = True
                results.extend(execution.results)
                if execution.error:
                    error = True
            if info:
                infoobj.join()
            if error:
                print("error", file=sys.stderr)
                results = None
                raise Exception("error in one or more of the executors")
            else:
                results.sort(cmp=lambda a, b: cmp(a[0], b[0]))
                results = [k[1] for k in results]
                # print "bla", results
                if flatten:
                    flatresults = []
                    for result in results:
                        flatresults.extend(result)
                    results = flatresults
            return results
        return execute
    return wrapper


if __name__ == "__main__":
    @timed
    @parallelize(cores=6, fork=True, flatten=True, text="common argument")
    def testprime(from_nr, to_nr, text=""):
        # print text, from_nr, to_nr
        primes = []
        from_nr = max(from_nr, 2)
        for p in range(from_nr, to_nr):
            isprime = True
            time.sleep(1)

            for i in range(2, p):
                if p % i == 0:
                    isprime = False
                    break
            if isprime:
                primes.append(p)
        return primes

    testnumbers = list(range(0, 10001, 100))
    from_nrs = testnumbers[:-1]
    to_nrs = testnumbers[1:]
    results = testprime(from_nrs, to_nrs)
    # print results
    # print countcores()
